// some windows typedefs
#typedef BOOL     int
#typedef HWND     unsigned int
#typedef DWORD    unsigned int
#typedef HDC      unsigned int
#typedef HFONT    unsigned int
#typedef FARPROC  unsigned int
#typedef COLORREF unsigned int
#typedef HCURSOR  unsigned int

// private flags in m_ctlflags
#define LISTBUTTON_FLAG_BTNVISIBLE  0x0001 // - set in OnPaint ~
#define LISTBUTTON_FLAG_LBXVISIBLE  0x0002
#define LISTBUTTON_FLAG_EDITVISIBLE 0x0004
#define LISTBUTTON_FLAG_BTNPRESSED  0x0200 // same as DFCS_PUSHED
#define LISTBUTTON_FLAG_BTNHOT      0x1000 // same as DFCS_HOT
#define PROPLIST_RESIZE_DISABLED    0x0008
#define PROPLIST_RESIZE_INACTION    0x0010
// HIWORD(m_ctlflags) = item.selected.RECT.client.top offset in pixels
// internal settings:
// - max visible items in listbox is 5
// - default column.0 width = 80
// - hex item must begin with '0x' (to convert editable hex string to int)

// todo:
// - item tooltips
// - validate editable items for selected format (int/hex) ?
// - capture SetFont
// - bold the selected items ?

// items storage
struct PROPITEM
{
	PROPITEM *next;
	string *title;
	DWORD   flags;     // 0xFF:type, 0xFF00:flags, 0xFFFF0000:string index for list
	DWORD   listitems; // number of strings in .data list - used for saving speed
	void   *data;      // bool, string or string-list in form string,0,string,0, ..., 0,0
}


class CPropertyListView : CControl
{
	// derieved standard methods
	declare virtual Create(int l,int t,int w,int h,int style,int exstyle,string title,CWindow *parent),int;
	declare virtual OnPaint(),int; // move/resize/hide edit/listbox, draw items
	declare virtual WndProc(unsigned int message, unsigned int wparam, unsigned int lparam),int;
	declare virtual OnVScroll(int nCommand,int nPos,int nID),int;
	declare virtual OnControl(int nID, int nNotifyCode, unsigned int hControl),int;
	declare virtual OnDestroy(),int; // kill listbox,edit and free items
	declare virtual OnKeyDown(unsigned int nChar, INT nRepCnt, unsigned int nFlags),int;
	declare virtual OnLButtonDown(int x,int y, int flags),int;
	declare virtual OnLButtonUp(int x,int y,int flags),int;
	declare virtual OnSize(int nType,int cx,int cy),int;
	declare virtual OnMouseMove(int x,int y,int flags),int;
	declare virtual OnLButtonDblClk(int x,int y,int flags),int;
	declare virtual OnTimer(int nIDEvent),int;

public:
	// events fired without messages
	declare virtual OnSelectionChanged(int item);
	declare virtual OnDropdown(int item),BOOL; // return true to show listbox

	declare Add(string *name,DWORD type,void *value);
	declare SetListItemString(int item,int stringIndex,string *str),BOOL;
	declare DeleteAllItems(); // called in OnDestroy
	declare EnableColumnResizing(BOOL resizable);
	declare EnableItem(int item,BOOL enabled);
	declare SetColumnWidth(DWORD width),DWORD;
	declare GetColumnWidth(),DWORD;
	declare RemoveSelection();
	declare SetSelected(int item),BOOL;
	declare GetItemsPerPage(),int;
	declare GetTopIndex(),int;
	declare GetItemTitle(int item),string *;
	declare GetSubitemText(int item),string *; // copy it immediatelly if item type is bool/int/uint/hex! Next paint cycle can change the string
	declare GetIntValue(int item),int; // bool,int,uint,hex. If item type is list, it tries to convert the string to integer

private:
	// internal methods
	declare OnListBoxSelection(int index);
	declare OnKillFocus(HWND hwndGotFocus),int;
	declare UpdateVScroll(); // SetScrollPos()
	declare GetDropdownButtonRect(RECT *rc);
	declare GetItemPtr(int item),PROPITEM *;
	declare DrawListButton(HDC dc,opt RECT *rc=0); // OnPaint, OnMouseMove, OnLButtonDown/Up ...
	declare SetSizeCursor(int columnW,int mouseX),BOOL;
	declare SetWindowPos2(HWND hwnd,int pos,int x,int y,int w,int h,int flags); // moves the listbox using screen coordinates
	declare ShowDropdown(); // calls OnDropdown, ShowListbox and FillListbox
	declare ShowListbox();
	declare FillListbox();
	declare HideListbox();
	declare ShowEdit();
	declare HideEdit();

//	HFONT    m_hFont2;     // default font is DEFAULT_GUI_FONT (not deleted int OnDestroy)
	HWND     m_hEdit;      // HWND edit
	HWND     m_hListbox;   // HWND listbox
	FARPROC  m_lpfnedit;   // subclassed edit
	FARPROC  m_lpfnlist;   // subclassed listbox
	PROPITEM *m_pItems;    // items storage
public:
	int      m_topindex;   // first visible item
	int      m_itemcnt;    // number of items
	int      m_col0wid;    // width of first column
	// m_fg and m_bg are handled too
	COLORREF m_gridcol;    // color of grid lines
	COLORREF m_selbg;      // color of selected item background
	int      m_itemSpacing;
	int      m_selected;   // selected item index
	byte     m_szTemp[32]; // space for string formatting (OnPaint), and for GetSubitemText, if item type is bool,int,uint or hex (32 bytes is for unicode version)
	int      m_ctlflags;   // internal flags, see bellow
	HCURSOR  m_hCursorC;   // custom cursor for resizing first column
}
#define ITEM_TYPE_BOOL     0x0001 // types and flags for Add()
#define ITEM_TYPE_INT      0x0002
#define ITEM_TYPE_UINT     0x0003
#define ITEM_TYPE_HEX      0x0004 // display integer as 0x....
#define ITEM_TYPE_STRING   0x0005
#define ITEM_TYPE_LIST     0x0006

#define ITEM_FLAG_DISABLED 0x0100
#define ITEM_FLAG_READONLY 0x0200
#define ITEM_FLAG_SELECTED 0x0400
#define ITEM_FLAG_EDITABLE 0x0800 // if set:uses edit control to change the value (it is automatically set for types:int,uint,hex and string)

// m_ctlflags - hiword: selected item offset from client.top
#define LISTBUTTON_FLAG_BTNVISIBLE  0x0001 // for cropping subitem text (set in OnPaint)
#define LISTBUTTON_FLAG_LBXVISIBLE  0x0002 // listbox is visible (for methods Show/Hide Listbox)
#define LISTBUTTON_FLAG_EDITVISIBLE 0x0004 // edit is visible (for methods Show/Hide Edit)
#define LISTBUTTON_FLAG_BTNPRESSED  0x0200 // dropdown button is pressed
#define LISTBUTTON_FLAG_BTNHOT      0x1000 // dropdown button is under the mouse pointer
#define PROPLIST_RESIZE_DISABLED    0x0008 // column resizing is disabled
#define PROPLIST_RESIZE_INACTION    0x0010 // user is resizing the first column

/*
	keyboard:
	- up/down changes the selection
	- F4 shows/hides listbox for list-items
	- RETURN in editable items moves the current selection to next item
	- any changes in editable items are accepted

	mouse:
	- wheel scrolls the control up/down
	- double-click on list-items changes displayed string to next, or first if current string was the last
	- over column divider: allows to resize the column.
	- over list-items button: changes the cursor and applies some visual styles to the button

*/